import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import math
import time
# self-define
# import lsh.lsh_mips as lsh

class GraphAttentionLayer(nn.Module):
    """
    Simple GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, dropout, alpha, concat=True):
        super(GraphAttentionLayer, self).__init__()
        self.dropout = dropout
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.concat = concat

        self.W = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        nn.init.xavier_uniform_(self.W.data, gain=1.414) # one special init method,by  Bengio.
        self.a = nn.Parameter(torch.zeros(size=(2*out_features, 1)))
        nn.init.xavier_uniform_(self.a.data, gain=1.414)

        self.leakyrelu = nn.LeakyReLU(self.alpha)

    def forward(self, input, adj):
        h = torch.mm(input, self.W) # mm -> matrix multiplication 
        N = h.size()[0]

        # h.repeat(x,y,z) means expand x,y,z 倍 in each dimension respectivly.
        a_input = torch.cat([h.repeat(1, N).view(N * N, -1), h.repeat(N, 1)], dim=1).view(N, -1, 2 * self.out_features)
        # print("a_input shape:{}".format(a_input.shape))
        # [N*N,2*out_features]
        e = self.leakyrelu(torch.matmul(a_input, self.a).squeeze(2))
        # [N,N]
        # torch.matmul -> multiplication in tensors. (4,5,6)*(2,5) -> (4,2,6)
        # torch.squeeze() needs take the size 1 in tensor dimension out. : if the 2th dimension is size 1,then take it out.
        
        # e.shape ->N,N

        zero_vec = -9e15*torch.ones_like(e)
        attention = torch.where(adj > 0, e, zero_vec) # which means if here exists a edge,use e' element,else: use zero_vec' element(not zero beacuse softmax next).
        attention = F.softmax(attention, dim=1)
        attention = F.dropout(attention, self.dropout, training=self.training)
        h_prime = torch.matmul(attention, h)

        if self.concat:
            return F.elu(h_prime) # avtivate or not.
        else:
            return h_prime

    def __repr__(self): # representatioin
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


class LSHAttentionLayer(nn.Module):
    """
    Simple GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, dropout, alpha, concat=True,seed=3933,cuda=True,bucket=4):
        super(LSHAttentionLayer, self).__init__()
        self.dropout = dropout
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.concat = concat
        self.seed=seed
        self.cuda=cuda
        self.bucket=bucket
        
        # some infomation for LSH config
        self.nHashTable=1
        self.nbuckets=4

        self.kW = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        nn.init.xavier_uniform_(self.kW.data, gain=1.414) # one special init method,by  Bengio.
        
        self.vW = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        nn.init.xavier_uniform_(self.vW.data, gain=1.414) # one special init method,by  Bengio.
        
        # self.a = nn.Parameter(torch.zeros(size=(2*out_features, 1)))
        # nn.init.xavier_uniform_(self.a.data, gain=1.414)

        self.leakyrelu = nn.LeakyReLU(self.alpha)

    def forward(self, input, adj):
        # t1=time.time()
        kh = torch.mm(input, self.kW) # mm -> matrix multiplication 
        N,hiddenSize = kh.shape
        
        # t2=time.time()
        vh =torch.mm(input,self.vW)
        # t3=time.time()
        # h.repeat(x,y,z) means expand x,y,z 倍 in each dimension respectivly.
        # shape of a_input: [N,N,2*8]
        
        # t4=time.time()         
        nHashTable=1
        bucketSize=self.bucket
        if self.cuda:                                                  
            rotations_rands=torch.randn(nHashTable,hiddenSize,bucketSize//2).cuda() #[nhashtable,hiddenSize,bucketSize//2]
        else:
            rotations_rands=torch.randn(nHashTable,hiddenSize,bucketSize//2)
        # t5=time.time()
        rota_vectors=torch.matmul(1.0*kh,rotations_rands)                #[nhashtable,N,bucketsize//2]
        # t6=time.time()
        rota_vectors=torch.cat([rota_vectors,-rota_vectors],-1)           #[nhashtable,N,bucketsize]
        buckets=torch.argmax(rota_vectors,dim=2)                       #[nhashtable,N]
        # 此时返回的是索引，这个索引可以理解成是bucket，代表了不同节点的bucket
        # print(buckets)
        # t7=time.time()
        #-------------------------------need to be optimalize-------------------------
        values,indexes=buckets.sort(dim =1)                              #[nhashtable]
        # 这里的value就是上面的bucket，这里的indexes就是排序之后的value在之前位置的index
        #-----------------------------------------------------------------------------
        # t8=time.time()
        # to x class
        values=values.tolist()[0]
        indexes=indexes.tolist()[0]
        # print(indexes)
        # t9=time.time()
        tempList=[]
        indexList=[]

        # print(values)

        for i in range(bucketSize):
            if i in values:
                tempList.append(values.index(i)) # 这个循环的意思是：对于每个bucket（也就是i），都需要找到相对应的开始index。
        # t11=time.time()
        for i in range(len(tempList)):
            if i != len(tempList)-1:
                # print(i)
                indexList.append(indexes[tempList[i]:tempList[i+1]])
            else:
                indexList.append(indexes[tempList[i]:])
        # print(indexList)
        # t12=time.time()
        del tempList        
        # t13=time.time()
        if self.cuda:
            K=-9e15*torch.ones((N,N)).cuda()
        else:
            K=-9e15*torch.ones((N,N))
        #print(indexList)
        # print('=='*10+'\n'+str(N)+'\n\n')
        for x in indexList:
            # print(len(x))
            
            K[torch.tensor(x).repeat(len(x)),torch.tensor(x).repeat(len(x),1).T.reshape(1,-1).view(-1)]=torch.mm(kh[x,:],kh[x,:].T).view(-1)/math.sqrt(hiddenSize)
        # t14=time.time()
        # print(values,indexes)
        # rotations_rand=np.random.randn(hiddenSize,bucketSize//2,seed=self.seed)
        # rota_vectors=np.dot(h,rotations_rand)
        # rota_vectors=np.hstack([rota_vectors,-rota_vectors])
        # buckets=np.argmax(rota_vectors,axis=-1)
        
        # K=torch.mm(kh,kh.T)
        # K=K/(math.sqrt(hiddenSize))
        # t15=time.time()
        # a_input = torch.cat([h.repeat(1, N).view(N * N, -1), h.repeat(N, 1)], dim=1).view(N, -1, 2 * self.out_features)
        # print(type(a_input))
        # a_data=a_input.view(-1,2*self.out_features)
        # candidate,index=lsh.lsh_K_MIPS(K=10,L=2,datas=a_data.cpu().detach().numpy().tolist(),
        # querys=self.a.cpu().detach().numpy().tolist(),
        # num_neighbours=-1,rand_range=3,needInformation=False)
        # x,_=a_input.shape
        # result=torch.ones(a_input.shape)*(-9e15)
        # result[index]=candidate.reshape(N,N,len(index))
        # e = self.leakyrelu(torch.matmul(result, self.a).squeeze(2))
        # torch.matmul -> multiplication in tensors. (4,5,6)*(2,5) -> (4,2,6)
        # torch.squeeze() needs take the size 1 in tensor dimension out. : if the 2th dimension is size 1,then take it out.
        
        # e.shape -> N,N/2

        if self.cuda:

            zero_vec = -9e15*torch.ones_like(K).cuda()
        else:
            zero_vec = -9e15*torch.ones_like(K)
        # print(adj.shape)
        # print(K.shape)
        attention = torch.where(adj > 0, K, zero_vec) # which means if here exists a edge,use e' element,else: use zero_vec' element(not zero beacuse softmax next).
        attention = F.softmax(attention, dim=1)
        attention = F.dropout(attention, self.dropout, training=self.training)
        h_prime = torch.matmul(attention,vh)

        # print(t2-t1)
        # print(t3-t2)
        # print(t4-t3)
        # print(t5-t4)
        # print(t7-t6)
        # print(t8-t7)
        # print(t9-t8)
        # print(t11-t9)
        # print(t12-t11)
        # print(t13-t12)
        # print(t14-t13)
        # print(t15-t14)
        # # print(t16-t15)
        # # print(t17-t16)

        if self.concat:
            return F.elu(h_prime) # avtivate or not.
        else:
            return h_prime

    def __repr__(self): # representatioin
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'



from graphviz import Digraph

import torch
import models

def make_dot(var, params):
    """ Produces Graphviz representation of PyTorch autograd graph
    
    Blue nodes are the Variables that require grad, orange are Tensors
    saved for backward in torch.autograd.Function
    
    Args:
        var: output Variable
        params: dict of (name, Variable) to add names to node that
            require grad (TODO: make optional)
    """
    param_map = {id(v): k for k, v in params.items()}
    print(param_map)
    
    node_attr = dict(style='filled',
                     shape='box',
                     align='left',
                     fontsize='12',
                     ranksep='0.1',
                     height='0.2')
    dot = Digraph(node_attr=node_attr, graph_attr=dict(size="12,12"))
    seen = set()
    
    def size_to_str(size):
        return '('+(', ').join(['%d'% v for v in size])+')'

    def add_nodes(var):
        if var not in seen:
            if torch.is_tensor(var):
                dot.node(str(id(var)), size_to_str(var.size()), fillcolor='orange')
            elif hasattr(var, 'variable'):
                u = var.variable
                node_name = '%s\n %s' % (param_map.get(id(u)), size_to_str(u.size()))
                dot.node(str(id(var)), node_name, fillcolor='lightblue')
            else:
                dot.node(str(id(var)), str(type(var).__name__))
            seen.add(var)
            if hasattr(var, 'next_functions'):
                for u in var.next_functions:
                    if u[0] is not None:
                        dot.edge(str(id(u[0])), str(id(var)))
                        add_nodes(u[0])
            if hasattr(var, 'saved_tensors'):
                for t in var.saved_tensors:
                    dot.edge(str(id(t)), str(id(var)))
                    add_nodes(t)
    add_nodes(var.grad_fn)
    return dot

inputs = torch.randn(100, 50).cuda()
adj = torch.randn(100, 100).cuda()
model = models.SpGAT(50, 8, 7, 0.5, 0.01, 3)
model = model.cuda()
y = model(inputs, adj)

g = make_dot(y, model.state_dict())
g.view()



import torch
import torch.nn as nn
import torch.nn.functional as F
from  fastGAT/layer import GraphAttentionLayer,LSHAttentionLayer


class GAT(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, alpha, nheads):
        """Dense version of GAT."""
        super(GAT, self).__init__()
        self.dropout = dropout

        self.attentions = [GraphAttentionLayer(nfeat, nhid, dropout=dropout, alpha=alpha, concat=True) for _ in range(nheads)]
        for i, attention in enumerate(self.attentions):
            self.add_module('attention_{}'.format(i), attention)

        self.out_att = GraphAttentionLayer(nhid * nheads, nclass, dropout=dropout, alpha=alpha, concat=False)

    def forward(self, x, adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x = torch.cat([att(x, adj) for att in self.attentions], dim=1)
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.elu(self.out_att(x, adj))
        return F.log_softmax(x, dim=1)

class FastGAT(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, alpha, nheads,bucket=4,cuda=True):
        super(FastGAT, self).__init__()
        self.dropout = dropout
        self.layernorm=nn.LayerNorm(nhid*nheads)

        self.attention1 = [LSHAttentionLayer(nfeat, nhid, dropout=dropout, alpha=alpha,bucket=bucket,cuda=cuda, concat=True) for _ in range(nheads)]
        for i, attention in enumerate(self.attention1):
            self.add_module('attention1_{}'.format(i), attention)
        self.inter_att=LSHAttentionLayer(nhid*nheads,nfeat,dropout=dropout,alpha=alpha,concat=False,bucket=bucket,cuda=cuda)
        #'''
        self.attention2 = [LSHAttentionLayer(nhid*nheads, nhid, dropout=dropout, alpha=alpha,bucket=bucket,cuda=cuda, concat=True) for _ in range(nheads)]
        for i, attention in enumerate(self.attention2):
            self.add_module('attention2_{}'.format(i), attention)
        #'''
        self.out_att = LSHAttentionLayer(nhid * nheads, nclass, dropout=dropout, alpha=alpha, concat=False,bucket=bucket,cuda=cuda)

    def forward(self,x,adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x1 = torch.cat([att(x, adj) for att in self.attention1], dim=1)
        x1 = F.dropout(x1, self.dropout, training=self.training)
        #print("x:{0},\n,x1:{1}\n".format(x.shape,x1.shape))
        
        #lone=self.layernorm(x1+x)
        #x2 = F.elu(self.inter_att(lone,adj))
        #print(x2.shape)
        #print(x.shape)
        #x2 = F.dropout(x2, self.dropout, training=self.training)
        #x2 = torch.cat([att(x2,adj) for att in self.attention2],dim=1)
        #x2 = F.dropout(x2, self.dropout, training=self.training)
        x = F.elu(self.out_att(x1, adj))
        return F.log_softmax(x, dim=1)





import setuptools

with open("README.md",'r') as f:
    descrip=f.read()

setuptools.setup(
    name="fastGAT",
    version="0.0.6",
    author='LiangZi@NEU',
    author_email='2273067585@qq.com',
    description='A Fast Graph ATtention networks with ALSH. ',
    long_description=descrip,
    long_description_content_type='text/markdown',
    url='',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],

)



import numpy as np
import scipy.sparse as sp
import torch


def encode_onehot(labels):
    '''
    generate one-hot label
    '''
    classes = set(labels)
    # np.identity :创建单位方阵
    classes_dict = {c: np.identity(len(classes))[i, :] for i, c in enumerate(classes)}
    labels_onehot = np.array(list(map(classes_dict.get, labels)), dtype=np.int32)
    return labels_onehot


def load_data(path="./data/cora/cora"):
    """Load citation network dataset (cora only for now)"""
    print('Loading dataset...')

    idx_features_labels = np.genfromtxt("{}.content".format(path), dtype=np.dtype(str))
    features = sp.csr_matrix(idx_features_labels[:, 1:-1], dtype=np.float32)
    labels = encode_onehot(idx_features_labels[:, -1])

    # build graph
    idx = np.array(idx_features_labels[:, 0], dtype=np.int32)
    idx_map = {j: i for i, j in enumerate(idx)}
    edges_unordered = np.genfromtxt("{}.cites".format(path), dtype=np.int32)
    edges = np.array(list(map(idx_map.get, edges_unordered.flatten())), dtype=np.int32).reshape(edges_unordered.shape)
    adj = sp.coo_matrix((np.ones(edges.shape[0]), (edges[:, 0], edges[:, 1])), shape=(labels.shape[0], labels.shape[0]), dtype=np.float32)

    # build symmetric adjacency matrix
    adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)

    features = normalize_features(features)
    adj = normalize_adj(adj + sp.eye(adj.shape[0]))

    idx_train = range(140)
    idx_val = range(200, 500)
    idx_test = range(500, 1500)

    adj = torch.FloatTensor(np.array(adj.todense()))
    features = torch.FloatTensor(np.array(features.todense()))
    labels = torch.LongTensor(np.where(labels)[1])

    idx_train = torch.LongTensor(idx_train)
    idx_val = torch.LongTensor(idx_val)
    idx_test = torch.LongTensor(idx_test)

    return adj, features, labels, idx_train, idx_val, idx_test


def normalize_adj(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv_sqrt = np.power(rowsum, -0.5).flatten()
    r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
    r_mat_inv_sqrt = sp.diags(r_inv_sqrt)
    return mx.dot(r_mat_inv_sqrt).transpose().dot(r_mat_inv_sqrt)


def normalize_features(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx


def accuracy(output, labels):
    preds = output.max(1)[1].type_as(labels)
    correct = preds.eq(labels).double()
    correct = correct.sum()
    return correct / len(labels)




import os
import glob
import time
import random
import argparse
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.autograd import Variable

from fastGAT/utils import load_data, accuracy
from fastGAT/model import GAT,FastGAT


class exe(): 
    def __init__(self,save_model=True,datasetpath='./data/cora',no_cuda=False,bucket=4,fastmode=False,seed=3933,epochs=200,lr=0.005,weight_delay=5e-4,hidden=8,nb_heads=8,dropout=0.6,alpha=0.2,patience=100,fastGAT=1):


        self.fastmode=fastmode
        self.seed=seed
        self.epochs=epochs
        # self.lr=lr
        # self.weight_delay=weight_delay
        self.patience=patience
        self.fastGAT=fastGAT
        self.save_model=save_model
        self.datasetpath=datasetpath
        self.bucket=bucket

        self.cuda=not no_cuda and torch.cuda.is_available()

        self.best_epoch_during_training=-1

        self.save_name='.'+str(self.fastGAT)+'.pkl'

        # init seed.
        random.seed(self.seed)
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        if self.cuda:
            torch.cuda.manual_seed(self.seed)

        # load data
        adj, features, labels, self.idx_train, self.idx_val, self.idx_test = load_data(datasetpath)        

        # load model
        if fastGAT==0:
            # Model and optimizer
            self.model = GAT(nfeat=features.shape[1], 
                        nhid=hidden, 
                        nclass=int(labels.max()) + 1, 
                        dropout=dropout, 
                        nheads=nb_heads, 
                        alpha=alpha)
            self.optimizer = optim.Adam(model.parameters(), 
                                lr=lr, 
                                weight_decay=weight_decay)
        else:
            # Model and optimizer
            self.model = FastGAT(nfeat=features.shape[1], 
                        nhid=hidden, 
                        nclass=int(labels.max()) + 1, 
                        dropout=dropout, 
                        nheads=nb_heads,
                        bucket=self.bucket, 
                        alpha=alpha)
            self.optimizer = optim.Adam(self.model.parameters(), 
                                lr=lr, 
                                weight_decay=weight_delay)


        if self.cuda:
            self.model.cuda()
            features = features.cuda()
            adj = adj.cuda()
            labels = labels.cuda()
            self.idx_train = self.idx_train.cuda()
            self.idx_val = self.idx_val.cuda()
            self.idx_test = self.idx_test.cuda()

        self.features, self.adj, self.labels = Variable(features), Variable(adj), Variable(labels)    


    def tra_val(self):
        model=self.model
        optimizer=self.optimizer
        # Train model
        t_total = time.time()
        loss_values = []
        bad_counter = 0
        best = self.epochs + 1
        best_epoch = 0
        save_name=self.save_name

        #=====================================per epoch===========================================
        for epoch in range(self.epochs):
            t = time.time()
            # model.train()
            optimizer.zero_grad()
            output = model(self.features, self.adj)
            loss_train = F.nll_loss(output[self.idx_train], self.labels[self.idx_train])
            acc_train = accuracy(output[self.idx_train], self.labels[self.idx_train])
            loss_train.backward()
            optimizer.step()

            if not self.fastmode:
                # Evaluate validation set performance separately,
                # deactivates dropout during validation run.
                model.eval()
                output = model(self.features, self.adj)

            loss_val = F.nll_loss(output[self.idx_val], self.labels[self.idx_val])
            acc_val = accuracy(output[self.idx_val], self.labels[self.idx_val])
            print('Epoch: {:04d}'.format(epoch+1),
                'loss_train: {:.4f}'.format(loss_train.data.item()),
                'acc_train: {:.4f}'.format(acc_train.data.item()),
                'loss_val: {:.4f}'.format(loss_val.data.item()),
                'acc_val: {:.4f}'.format(acc_val.data.item()),
                'time: {:.4f}s'.format(time.time() - t))

            loss_values.append(loss_val.data.item())

            torch.save(model.state_dict(), ('{}'+save_name).format(epoch))
            if loss_values[-1] < best:
                best = loss_values[-1]
                best_epoch = epoch
                bad_counter = 0
            else:
                bad_counter += 1

            if bad_counter == self.patience:
                break

            files = glob.glob('*'+save_name)
            for file in files:
                epoch_nb = int(file.split('.')[0])
                if epoch_nb < best_epoch:
                    os.remove(file)

        files = glob.glob('*'+save_name)
        for file in files:
            epoch_nb = int(file.split('.')[0])
            if epoch_nb > best_epoch:
                os.remove(file)

        print("Training Finished.")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print('Loading {}th epoch'.format(best_epoch))

        self.best_epoch_during_training=best_epoch
        
        print('OK.')



    def test(self):


        print("make sure you have train the model. or test will be failed.")

        # Reload best model
        model=self.model
        model.load_state_dict(torch.load(('{}'+self.save_name).format(self.best_epoch_during_training)))

        # Testing
        model.eval()
        output = model(self.features, self.adj)
        loss_test = F.nll_loss(output[self.idx_test], self.labels[self.idx_test])
        acc_test = accuracy(output[self.idx_test], self.labels[self.idx_test])
        print("Test set results:",
            "loss= {}".format(loss_test.item()),
            "accuracy= {:.4f}".format(acc_test.item()))
        print("done.")



from fastGAT/fastGAT import exe

procedure=exe(save_model=True,datasetpath='./data/cora/cora',no_cuda=False,
            fastmode=False,seed=3933,epochs=200,lr=0.005,weight_delay=5e-4,hidden=8,nb_heads=4,
            bucket=2,dropout=0.6,alpha=0.2,patience=100,fastGAT=1)
    
procedure.tra_val()

procedure.test()